/* -*-C++-*-
 * ###################################################################
 *  Cpptcl - connecting C++ with Tcl
 * 
 *  FILE: "cpptcl_matrix.cc"
 *                                    created: 18/9/97 {11:24:31 am} 
 *                                last update: 26/1/98 {5:47:42 pm} 
 *  Author: Vince Darley
 *  E-mail: <darley@fas.harvard.edu>
 *    mail: Division of Engineering and Applied Sciences, Harvard University
 *          Oxford Street, Cambridge MA 02138, USA
 *     www: <http://www.fas.harvard.edu/~darley/>
 *  
 * ========================================================================
 *               Copyright (c) 1997 Vince Darley
 * ========================================================================
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that the copyright notice and warranty disclaimer appear in
 * supporting documentation.
 * 
 * Vince Darley disclaims all warranties with regard to this software,
 * including all implied warranties of merchantability and fitness.  In
 * no event shall Vince Darley be liable for any special, indirect or
 * consequential damages or any damages whatsoever resulting from loss of
 * use, data or profits, whether in an action of contract, negligence or
 * other tortuous action, arising out of or in connection with the use or
 * performance of this software.
 * ========================================================================
 *  See header file for further information
 * ###################################################################
 */

#include "cpptcl_matrix.h"
#include "tclMatrix.h"

/// Here's the C++ package init procedure
int Cpptcl_InitFunction(Cppmatrix_Init) {
	interp.PkgRequire("Cpptcl");
	// Complex macro which relies on the stream being called 'interp'
	// It declares a new C++ class and from what it derives.
    Cpptcl_Object(cpptcl_matrix, tcl_object);
	Tcl_RegisterMatrixType(&Cppmatrix_GetMatrixPtr);
	
    interp.PkgProvide("Cppmatrix", "0.1");
    return TCL_OK;
}

extern "C" tclMatrix* Cppmatrix_GetMatrixPtr(Tcl_Interp *interp, 
											const char *name){
	tcl_base* o = Cpptcl_getObjectByName(interp,name);
	if(o) {
		cpptcl_matrix* m = ckcast(o,cpptcl_matrix);
		if(m) { return &m->mat;}
	}
	return (tclMatrix*)NULL;
}

#define I_3D(i,j,k)	k + mat.n[2] * (I_2D(i,j))
#define I_2D(i,j)	j + mat.n[1] * (I_1D(i))
#define I_1D(i)		i

cpptcl_matrix::cpptcl_matrix(tcl_args& arg)
	:tcl_object(arg),mat()
{
    mat.fdata = NULL;
    mat.idata = NULL;
    mat.name = NULL;
    mat.dim = 0;
    mat.len = 1;
    mat.tracing = 0;
	mat.put = 0;
	mat.get = 0;
    for (int i = 0; i < MAX_ARRAY_DIM; i++)
		mat.n[i] = 1;

	if(arg("","","i") == "int") {
		mat.type = TYPE_INT;
		// only needed for backwards compatibility
	} else if(arg("","","f") == "float") { 
		mat.type = TYPE_FLOAT;
		// only needed for backwards compatibility
	} else {
		arg.no_match();
		return;
	}
	for(int ii = 0;ii < 4; ii++) {
		if(arg.peek_is("=")) {
			// initialiser
			break;
		}
		if(ii <3) {
			arg >> optional >> mat.n[ii];
			if(mat.n[ii] > 1) 
				mat.dim++;
			mat.len *= mat.n[ii];
		}
	}
	if(mat.dim == 0) mat.dim = 1;
	
	/* Allocate space for data */

    switch (mat.type) {
	  case TYPE_FLOAT:
		mat.fdata = new Mat_float[mat.len];
		if (mat.fdata == NULL) {
			tcl_ << "Couldn't allocate " << mat.len << " elements for matrix.";	
			arg.signal_error(tcl_args::Cpp_constructor);
			return;
		}
		for (int i = 0; i < mat.len; i++)
	    	mat.fdata[i] = 0.0;
		break;

      case TYPE_INT:
		mat.idata = new Mat_int[mat.len];
		if (mat.idata == NULL) {
			tcl_ << "Couldn't allocate " << mat.len << " elements for matrix.";	
			arg.signal_error(tcl_args::Cpp_constructor);
			return;
		}
		for (int i = 0; i < mat.len; i++)
	    	mat.idata[i] = 0;
		break;
    }
}

int cpptcl_matrix::parse_tcl_command(tcl_args& arg){
	if(arg("?xmin? ?xmax? ?ymin....?","return a range of matrix elements")=="dump") {
		int nmin[MAX_ARRAY_DIM], nmax[MAX_ARRAY_DIM];
		for (int i = 0; i < MAX_ARRAY_DIM; i++) {
			nmin[i] = 0;
			nmax[i] = mat.n[i]-1;
		}
		for (int i = 0; i < MAX_ARRAY_DIM; i++) {
			arg >> optional >> nmin[i] >> optional >> nmax[i];
		}
		arg >> done;
		NO_EXCEPTIONS(arg);
		for (int i = 0; i < MAX_ARRAY_DIM; i++) {
			if(nmin[i] < 0 || nmax[i] > mat.n[i]-1) {
				tcl_ << "coordinate " << i << " out of range [0," 
					 << mat.n[i]-1 << "]" << tcl_error;
				return tcl_;
			}
		}
		for (int i = nmin[0]; i <= nmax[0]; i++) {
		    for (int j = nmin[1]; j <= nmax[1]; j++) {
				for (int k = nmin[2]; k <= nmax[2]; k++) {
					switch (mat.type) {
					  case TYPE_FLOAT:
						tcl_ << mat.fdata[I_3D(i,j,k)] << " ";
						break;
				
					  case TYPE_INT:
						tcl_ << mat.idata[I_3D(i,j,k)] << " ";
						break;
					}
				}
				if (mat.dim > 2)
					tcl_ << endl;
			}
			if (mat.dim > 1)
				tcl_ << endl;
		}
		tcl_ << result;
		return tcl_;
	} else if(arg("?xmin? ?xmax? ?yvalue? ?zvalue?","return a range of matrix elements")=="dumpbinary") {
		int nmin[MAX_ARRAY_DIM], nmax[MAX_ARRAY_DIM];
		for (int i = 0; i < MAX_ARRAY_DIM; i++) {
			nmin[i] = 0;
			nmax[i] = mat.n[i]-1;
		}
		arg >> optional >> nmin[0] >> optional >> nmax[0];
		for (int i = 1; i < MAX_ARRAY_DIM; i++) {
			arg >> optional >> nmin[1] >> optional >> nmin[2];
		}
		arg >> done;
		NO_EXCEPTIONS(arg);
		nmax[1] = nmin[1];
		nmax[2] = nmin[2];
		for (int i = 0; i < MAX_ARRAY_DIM; i++) {
			if(nmin[i] < 0 || nmax[i] > mat.n[i]-1) {
				tcl_ << "coordinate " << i << " out of range [0," 
					 << mat.n[i]-1 << "]" << tcl_error;
				return tcl_;
			}
		}
		
		switch (mat.type) {
		  case TYPE_FLOAT:
			tcl_.binary_append(mat.fdata + I_3D(nmin[0],nmin[1],nmin[2]), 
				nmax[0]-nmin[0]+1);
			break;
		  case TYPE_INT:
			tcl_.binary_append(mat.idata + I_3D(nmin[0],nmin[1],nmin[2]), 
				nmax[0]-nmin[0]+1);
			break;
		}
		tcl_ << result;
		return tcl_;		
	} else if (arg("?length?","find minimum value")=="min") {
		int len = mat.len;
		arg >> optional >> len >> done;
		NO_EXCEPTIONS(arg);
		switch (mat.type) {
		case TYPE_FLOAT:{
			Mat_float min = mat.fdata[len-1];
			while (len>0)
				if(mat.fdata[--len] < min) min = mat.fdata[len];
			tcl_ << min << result;
			break;
		}
		case TYPE_INT:{
			Mat_int min = mat.idata[len-1];
			while (len>0)
				if(mat.idata[--len] > min) min = mat.idata[len];
			tcl_ << min << result;
			break;
		}}
		return tcl_;
	} else if (arg("?length?","find maximum value")=="max") {
		int len = mat.len;
		arg >> optional >> len >> done;
		NO_EXCEPTIONS(arg);
		switch (mat.type) {
		case TYPE_FLOAT:{
			Mat_float max = mat.fdata[len-1];
			while (len>0)
				if(mat.fdata[--len] > max) max = mat.fdata[len];
			tcl_ << max << result;
			break;
		}
		case TYPE_INT:{
			Mat_int max = mat.idata[len-1];
			while (len>0)
				if(mat.idata[--len] > max) max = mat.idata[len];
			tcl_ << max << result;
			break;
		}}
		return tcl_;
	} else if (arg("","find matrix info")=="info") {
		arg >> done;
		NO_EXCEPTIONS(arg);
		for (int i = 0; i < mat.dim; i++)
			tcl_ << mat.n[i] << lappend;
		tcl_ << result;
		return tcl_;
	} else if (arg("","find matrix info")=="redim") {
		long newlen;
		arg >> newlen >> done;
		NO_EXCEPTIONS(arg);
		if (mat.dim != 1) {
			tcl_ << "can only redim a 1d matrix" << tcl_error;
			return tcl_;
		}
		if(newlen == mat.len) return tcl_;
		switch (mat.type) {
		  case TYPE_FLOAT: 
			{
			Mat_float* data = new Mat_float[newlen];
			if (data == NULL) {
				tcl_ << "redim failed!" << tcl_error;
				return tcl_;
			}
			long i;
			for (i = 0;i<mat.len && i<newlen;i++)
				data[i] = mat.fdata[i];
			delete [] mat.fdata;
			mat.fdata = data;
			for ( ; i < newlen; i++)
				mat.fdata[i] = 0.0;
			break;
			}
			
		  case TYPE_INT:
			{
			Mat_int* data = new Mat_int[newlen];
			if (data == NULL) {
				tcl_ << "redim failed!" << tcl_error;
				return tcl_;
			}
			long i;
			for (i = 0;i<mat.len && i<newlen;i++)
				data[i] = mat.idata[i];
			delete [] mat.idata;
			mat.idata = data;
			for ( ; i < newlen; i++)
				mat.idata[i] = 0.0;
			break;
			}
			
		}
		mat.n[0] = mat.len = newlen;
		return tcl_;
	} else if (arg("","unimplemented")=="filter") {
		arg >> done;
		NO_EXCEPTIONS(arg);
		return tcl_;
	} else if (arg("factor","increase length by given factor, filling in with suitable spacer values")=="interpolate") {
		int factor;
		arg >> factor >> done;
		NO_EXCEPTIONS(arg);
		if (mat.dim != 1) {
			tcl_ << "can only interpolate a 1d matrix" << tcl_error;
			return tcl_;
		}
		if(factor == 1) return tcl_;
		long newlen = mat.len*factor;
		switch (mat.type) {
		  case TYPE_FLOAT: 
			{
			Mat_float* data = new Mat_float[newlen];
			if (data == NULL) {
				tcl_ << "redim failed!" << tcl_error;
				return tcl_;
			}
			long i;
			for (i = 0;i<mat.len;i++) {
				data[i*factor] = mat.fdata[i];
				if(i != 0) {
					for(int j=1;j<factor;j++) {
						data[(i-1)*factor+j] = 
						  (data[i*factor]*j+data[(i-1)*factor]*(factor-j))/factor;
					}
				}
			}
			for(int j=1;j<factor;j++) {
				data[(i-1)*factor+j] = data[(i-1)*factor];
			}
			delete [] mat.fdata;
			mat.fdata = data;
			break;
			}
			
		  case TYPE_INT:
			{
			Mat_int* data = new Mat_int[newlen];
			if (data == NULL) {
				tcl_ << "redim failed!" << tcl_error;
				return tcl_;
			}
			long i;
			for (i = 0;i<mat.len && i<newlen;i++) {
				data[i*factor] = mat.idata[i];
				if(i != 0) {
					for(int j=1;j<factor;j++) {
						data[(i-1)*factor+j] = 
						  (data[i*factor]*j+data[(i-1)*factor]*(factor-j))/factor;
					}
				}
			}
			for(int j=1;j<factor;j++) {
				data[(i-1)*factor+j] = data[(i-1)*factor];
			}
			delete [] mat.idata;
			mat.idata = data;
			break;
			}
			
		}
		mat.n[0] = mat.len = newlen;
		return tcl_;
	} else if (arg("","unimplemented")=="scale") {
		float scale;
		arg >> scale >> done;
		NO_EXCEPTIONS(arg);
		switch (mat.type) {
		  case TYPE_FLOAT:
			for (long i = 0; i < mat.len; i++)
				mat.fdata[i] *= scale;
			break;
	
		  case TYPE_INT:
			for (long i = 0; i < mat.len; i++)
				mat.idata[i] *= scale;
			break;
		}
		return tcl_;
	} else if(arg("initialisation-list","fill matrix with values")=="=") {
		Tcl_Obj* obj = 0;
		arg >> obj >> done;
		NO_EXCEPTIONS(arg);
		return initialise(0,0,1,&obj);
	} else if(arg("index ... index ?= value?",
			"reads or sets matrix value(s)").is_a("int") || arg == "*") {
		arg >> backup;
		return put_get(arg);
	} else {
		return tcl_object::parse_tcl_command(arg);
	}
}

int cpptcl_matrix::put_get(tcl_args& arg) {
	int nmin[MAX_ARRAY_DIM], nmax[MAX_ARRAY_DIM];
	for (int i = 0; i < MAX_ARRAY_DIM; i++) {
		nmin[i] = 0;
		nmax[i] = mat.n[i]-1;
	}
	for (int i = 0; i < mat.dim; i++) {
		if (arg.peek_is_int()) {
			arg >> nmin[i];
			nmax[i] = nmin[i];
			if (nmin[i] < 0 || nmax[i] > mat.n[i]-1) {
				arg.signal_error(tcl_args::Syntax);
			}
		} else {
			arg -= "*";
			nmin[i] = 0;
			nmax[i] = mat.n[i]-1;
		}
	}
	bool put = false;
	if(!arg.empty()) {
		put = true;
		Mat_float tmp_f;
		Mat_int tmp_i;
		arg -= "=";
		if(arg.haveErr) {
			arg >> done;
			return TCL_ERROR;
		}	
	} else {
		arg >> done;
	}
	NO_EXCEPTIONS(arg);
	for (int i = nmin[0]; i <= nmax[0]; i++) {
		for (int j = nmin[1]; j <= nmax[1]; j++) {
			for (int k = nmin[2]; k <= nmax[2]; k++) {
				if (put) 
					switch (mat.type) {
					  case TYPE_FLOAT:
						arg >> mat.fdata[I_3D(i,j,k)];
						break;
					  case TYPE_INT:
						arg >> mat.idata[I_3D(i,j,k)];
						break;
					}
				else {
					switch (mat.type) {
					  case TYPE_FLOAT:
						tcl_ << mat.fdata[I_3D(i,j,k)] << lappend;
						break;
					  case TYPE_INT:
						tcl_ << mat.idata[I_3D(i,j,k)] << lappend;
						break;
					}
				}
			}
		}
	}
	if(put) {
		arg >> done;
	}
	NO_EXCEPTIONS(arg);		
	tcl_ << result;
	return tcl_;
}

cpptcl_matrix::~cpptcl_matrix(void){
	if(mat.fdata) {
		delete [] mat.fdata;
		mat.fdata = 0;
	}
	if(mat.idata) {
		delete [] mat.idata;
		mat.idata = 0;
	}
}

int cpptcl_matrix::initialise(int dim, int offs, int nargs, const Tcl_Obj** obj) {
    if (dim < mat.dim) {
		for (int i = 0; i < nargs; i++) {
			int split_list_len;
			Tcl_Obj** split_list;
		    if (Tcl_ListObjGetElements(NULL, (Tcl_Obj*)obj[i], 
									&split_list_len, &split_list) != TCL_OK) {
				return TCL_ERROR;
		    }
			if (split_list_len > mat.n[dim]) {
				return TCL_ERROR;
			}
			int newoffs = (dim > 0 ? offs*mat.n[dim - 1] + i : 0);
			if(initialise(dim + 1,newoffs,split_list_len,split_list) != TCL_OK) {
				return TCL_ERROR;
			}
		}
		return TCL_OK;
    }
	int newoffs = offs * mat.n[dim - 1];
    switch (mat.type) {
	  case TYPE_FLOAT:
		Mat_float *fptr = mat.fdata + newoffs;
		for (int i = 0; i < nargs; i++) {
			if(Tcl_GetDoubleFromObj(NULL, (Tcl_Obj*)obj[i], fptr+i) != TCL_OK) {
				return TCL_ERROR;
			}
		}
		break;
      case TYPE_INT:
		Mat_int *iptr = mat.idata + newoffs;
		for (int i = 0; i < nargs; i++) {
			if(Tcl_GetLongFromObj(NULL, (Tcl_Obj*)obj[i], iptr+i) != TCL_OK) {
				return TCL_ERROR;
			}
		}
		break;
    }

    return TCL_OK;
}

int cpptcl_matrix::parse_configuration_values(tcl_args& arg) {
	if(arg.peek_is("=")) {
		arg -= "=";
		Tcl_Obj* obj = 0;
		arg >> obj >> done;
		NO_EXCEPTIONS(arg);
		if(initialise(0,0,1,&obj) != TCL_OK) {
			arg.set_conversion_type("matrix initialisation list");
			arg.signal_error(tcl_args::Conversion);
			return TCL_ERROR;
		}
		//Tcl_DecrRefCount(obj);
	}
	return tcl_object::parse_configuration_values(arg);
}

int cpptcl_matrix::dim(void) const {return mat.dim;}

Cpptcl_Members(cpptcl_matrix) = {
	cppDatamemberfn("dim",&cpptcl_matrix::dim,0,int),
};

Cpptcl_IClass(cpptcl_matrix,"matrix",tcl_object);

